unit UpAudioSource;

interface

uses
  AudioSource,CosineTable,Math,Windows,SysUtils,Classes,Graphics;

const
  FilterSamples=$10000;

type
  TWaveType=(wtSinus,wtTriangle,wtSquare);

  TUpAudioSource=class(TMonoBufferAudioSource)
  private
    FWaves:Integer;
    FWaveTime:PFloats;
    FFilter:array[0..FilterSamples] of TFloat;
    FBase,FMultiplicator:TFloat;
    FWaveType:TWaveType;
    FSpeed,FTime:TFloat;
    FFilterMax,FFilterMedian,FFilterShape,FFilterWidth:TFloat;

    procedure SetWaves(const Value: Integer);

    procedure SetFilterMedian(const Value: TFloat);
    procedure SetFilterShape(const Value: TFloat);
    procedure SetFilterWidth(const Value: TFloat);
  protected
    function GetExportLength:TFloat;override;
    procedure Reset;override;

    function Filter(z:TFloat):TFloat;
    function FilterFunc(z:TFloat):TFloat;

    procedure AllocateWaves;
    procedure UnallocateWaves;

    procedure SampleFilter;
  public
    constructor Create;
    destructor Destroy;override;

    procedure Fill(M:PFloats;Samples,Frequency:Cardinal);override;

    procedure PaintTo(Canvas:TCanvas;Width,Height:Integer);

    property Waves:Integer read FWaves write SetWaves;
    property Base:TFloat read FBase write FBase;
    property Multiplicator:TFloat read FMultiplicator write FMultiplicator;
    property WaveType:TWaveType read FWaveType write FWaveType;

    property Speed:TFloat read FSpeed write FSpeed;
    property Progress:TFloat read FTime;

    property FilterMedian:TFloat read FFilterMedian write SetFilterMedian;
    property FilterShape:TFloat read FFilterShape write SetFilterShape;
    property FilterWidth:TFloat read FFilterWidth write SetFilterWidth;
  end;

implementation

function Triangle(X:TFloat):TFloat;
begin
  X:=Frac(X/(2*Pi));
  if X<0.5 then begin
    if X>0.25 then
      Result:=4*(0.5-X)
    else
      Result:=4*X;
  end else begin
    if X>0.75 then
      Result:=4*(X-1)
    else
      Result:=4*(0.5-X);
  end;
end;

function Square(X:TFloat):TFloat;
begin
  X:=Frac(X/(2*Pi));
  if X<0.5 then
    Result:=-1
  else
    Result:=1;
end;

{ TUpAudioSource }

procedure TUpAudioSource.AllocateWaves;
begin
  if FWaves<>0 then begin
    GetMem(FWaveTime,FWaves*SizeOf(TFloat));
    ZeroMemory(FWaveTime,FWaves*SizeOf(TFloat));
  end;
end;

constructor TUpAudioSource.Create;
begin
  inherited;
  FBase:=10;
  FMultiplicator:=Power(2,1/2);
  FSpeed:=0.1;
  FFilterMax:=0.5;
  FWaves:=Trunc(Ln(100000/FBase)/Ln(FMultiplicator));
  FFilterMedian:=0.4*FWaves;
  FFilterShape:=6;
  FFilterWidth:=10;
  AllocateWaves;
  SampleFilter;
end;

destructor TUpAudioSource.Destroy;
begin
  UnallocateWaves;
  inherited;
end;

procedure TUpAudioSource.Fill(M: PFloats; Samples, Frequency: Cardinal);
type
  TWaveFunc=function(X:TFloat):TFloat;
var
  a,b:Integer;
  f0,f,z,uf,dt,s:TFloat;
  w:TWaveFunc;
begin
  case FWaveType of
    wtTriangle:w:=Triangle;
    wtSquare:w:=Square;
  else
    w:=TCos;
  end;
  dt:=1/Frequency;
  uf:=Power(FMultiplicator,dt*FSpeed);
  f0:=FBase*Power(FMultiplicator,FTime);
  for a:=0 to Samples-1 do begin
    s:=0;
    f:=f0;
    z:=FTime;
    for b:=0 to FWaves-1 do begin
      s:=s+Filter(z)*w(2*Pi*FWaveTime[b]);
      FWaveTime[b]:=FWaveTime[b]+f*dt;
      z:=z+1;
      f:=f*FMultiplicator;
    end;
    M[a]:=s;
    f0:=f0*uf;
    FTime:=FTime+dt*FSpeed;
    while FTime>1 do begin
      FTime:=FTime-1;
      f0:=f0/FMultiplicator;
      for b:=FWaves-1 downto 1 do
        FWaveTime[b]:=FWaveTime[b-1];
    end;
    while FTime<0 do begin
      FTime:=FTime+1;
      f0:=f0*FMultiplicator;
      for b:=0 to FWaves-2 do
        FWaveTime[b]:=FWaveTime[b+1];
    end;
  end;
end;

function TUpAudioSource.Filter(z: TFloat): TFloat;
var
  n:Integer;
begin
  z:=z*FilterSamples/FWaves;
  n:=Trunc(z);
  Result:=FFilterMax*(FFilter[n]*(1+n-z)+FFilter[n+1]*(z-n));
end;

function TUpAudioSource.FilterFunc(z: TFloat): TFloat;
begin
  Result:=1/(1+Power(Abs(z-FFilterMedian)/FFilterWidth,FFilterShape));
end;

function TUpAudioSource.GetExportLength: TFloat;
begin
  Result:=1/FSpeed;
end;

procedure TUpAudioSource.PaintTo(Canvas: TCanvas; Width, Height: Integer);
const
  dx1=30;
  dx2=30;
  dy1=10;
  dy2=20;
var
  a,x,y:Integer;
  z,h:TFloat;

  procedure OutputText(x,y:Integer;s:string);
  var
    q:TSize;
  begin
    with Canvas do begin
      q:=TextExtent(s);
      TextOut(x-q.cx div 2,y-q.cy div 2,s);
    end;
  end;

begin
  with Canvas do begin
    Brush.Color:=clWhite;
    Pen.Color:=RGB(200,200,200);
    FillRect(Rect(0,0,Width,Height));
    Rectangle(dx1-1,dy1-1,Width-dx2+1,Height-dy2+1);
    Pen.Color:=clRed;
    for a:=dx1 to Width-dx2 do begin
      x:=a;
      z:=Filter((a-dx1)*FWaves/(Width-dx1-dx2));
      y:=Height-dy2-Round(z*(Height-dy1-dy2)/FFilterMax);
      if a=dx1 then
        MoveTo(x,y)
      else
        LineTo(x,y);
    end;
    Pen.Color:=clGray;
    OutputText(dx1,Height-dy2 div 2,FloatToStr(FBase)+' Hz');
    OutputText(Width-dx2,Height-dy2 div 2,FloatToStr(Round(FBase*Power(FMultiplicator,FWaves)))+' Hz');
    x:=dx1+Round(FFilterMedian*(Width-dx1-dx2)/FWaves);
    Pen.Style:=psDashDotDot;
    MoveTo(x,Height-dy2);
    LineTo(x,dy1);
    Pen.Style:=psSolid;
    OutputText(x,Height-dy2 div 2,FloatToStr(Round(FBase*Power(FMultiplicator,FFilterMedian)))+' Hz');
    Pen.Color:=clBlue;
    z:=FTime;
    for a:=1 to FWaves do begin
      h:=Filter(z);
      x:=dx1+Round(z*(Width-dx1-dx2)/FWaves);
      y:=Height-dy2-Round(h*(Height-dy1-dy2)/FFilterMax);
      MoveTo(x,Height-dy2);
      LineTo(x,y);
      z:=z+1;
    end;
  end;
end;

procedure TUpAudioSource.Reset;
begin
  UnallocateWaves;
  FTime:=0;
  AllocateWaves;
  SampleFilter;
end;

procedure TUpAudioSource.SampleFilter;
var
  a,b:Integer;
  z,s,m:TFloat;
begin
  for a:=0 to FilterSamples do begin
    z:=a*FWaves/FilterSamples;
    if z<1 then
      FFilter[a]:=z*FilterFunc(z)
    else begin
      if z>FWaves-1 then
        FFilter[a]:=FilterFunc(z)*(FWaves-z)
      else
        FFilter[a]:=FilterFunc(z);
    end;
  end;
  m:=1E-5;
  a:=Trunc(FilterSamples/FWaves);
  while a>=0 do begin
    s:=0;
    for b:=0 to FWaves-1 do
      s:=s+FFilter[(a+Trunc(b*FilterSamples/FWaves)) mod (FilterSamples+1)];
    if s>m then
      m:=s;
    Dec(a);
  end;
  FFilterMax:=1/m;
end;

procedure TUpAudioSource.SetFilterMedian(const Value: TFloat);
begin
  if not SameValue(FFilterMedian,Value) then begin
    FFilterMedian:=Value;
    SampleFilter;
  end;
end;

procedure TUpAudioSource.SetFilterShape(const Value: TFloat);
begin
  if not SameValue(FFilterShape,Value) then begin
    FFilterShape:=Value;
    SampleFilter;
  end;
end;

procedure TUpAudioSource.SetFilterWidth(const Value: TFloat);
begin
  if FFilterWidth<>Value then begin
    FFilterWidth:=Value;
    SampleFilter;
  end;
end;

procedure TUpAudioSource.SetWaves(const Value: Integer);
begin
  if FWaves<>Value then begin
    UnallocateWaves;
    FWaves:=Value;
    AllocateWaves;
    SampleFilter;
  end;
end;

procedure TUpAudioSource.UnallocateWaves;
begin
  if Assigned(FWaveTime) then begin
    FreeMem(FWaveTime);
    FWaveTime:=nil;
  end;
end;

end.

